
#include "stdafx.h"

#include "general.h"
#include "materials.h"
#include "RaVi.h"
#include "RaVi_Engine.h"

short int RV_AntiAlias = 0;
short int RV_Quality   = 1;
short int RV_Filter    = 0;

float RV_FoV = 0.5f;

inline unsigned long int BlendRGB2(unsigned long int a, unsigned long int b)
{
  return ((((a&0xFF0000)+(b&0xFF0000))>>1)&0xFF0000) | ((((a&0xFF00)+(b&0xFF00))>>1)&0xFF00) | ((((a&0xFF)+(b&0xFF))>>1)&0xFF);
}

inline unsigned long int BlendRGB3(unsigned long int a, unsigned long int b, unsigned long int c)
{
  return ((((a&0xFF0000)+(b&0xFF0000)+(c&0xFF0000))/3)&0xFF0000) | ((((a&0xFF00)+(b&0xFF00)+(c&0xFF00))/3)&0xFF00) | ((((a&0xFF)+(b&0xFF)+(c&0xFF))/3)&0xFF);
}

inline unsigned long int BlendRGB4(unsigned long int a, unsigned long int b, unsigned long int c, unsigned long int d)
{
  return ((((a&0xFF0000)+(b&0xFF0000)+(c&0xFF0000)+(d&0xFF0000))>>2)&0xFF0000) | ((((a&0xFF00)+(b&0xFF00)+(c&0xFF00)+(d&0xFF00))>>2)&0xFF00) | ((((a&0xFF)+(b&0xFF)+(c&0xFF)+(d&0xFF))>>2)&0xFF);
}

inline unsigned long int DiffRGB2(unsigned long int a, unsigned long int b)
{
  return abs((a>>16)-(b>>16)) + abs(((a>>8)&255)-((b>>8)&255)) + abs((a&255)-(b&255));
}

inline unsigned long int DiffRGB4(unsigned long int a, unsigned long int b, unsigned long int c, unsigned long int d)
{
  return abs((a>>16)-(b>>16)) + abs(((a>>8)&255)-((b>>8)&255)) + abs((a&255)-(b&255))
       + abs((c>>16)-(d>>16)) + abs(((c>>8)&255)-((d>>8)&255)) + abs((c&255)-(d&255))
       + abs((a>>16)-(d>>16)) + abs(((a>>8)&255)-((d>>8)&255)) + abs((a&255)-(d&255))
       + abs((c>>16)-(b>>16)) + abs(((c>>8)&255)-((b>>8)&255)) + abs((c&255)-(b&255));
}

unsigned long int TLin[340];
unsigned long int MLin[340];
unsigned long int BLin[340];

void DoFilter(byte *BMPbits, int Width, int Height)
{
  byte              *bitptr1;
  byte              *bitptr2;
  long unsigned int  ScLen;
  int                x,y;
  
  long unsigned int  cr,dg;

  long unsigned int *LBPt = TLin;
  long unsigned int *LBPm = MLin;
  long unsigned int *LBPb = BLin;

  ScLen= Width*4;
    
  bitptr1=BMPbits;
  // Initiate Linebuffer
  memcpy(LBPm+1,bitptr1,ScLen);
  bitptr2 = bitptr1; bitptr1 += ScLen;
  // extend borders
  LBPm[0]=LBPm[1]; LBPm[Width+1]=LBPm[Width];
  memcpy(LBPb,LBPm,4*(Width+2));

  for(y=0;y<Height;y++){
    // scroll line pointers
	if(LBPt==TLin){
      LBPt=MLin; LBPm=BLin; LBPb=TLin;
	} else if(LBPt==MLin){
      LBPt=BLin; LBPm=TLin; LBPb=MLin;
	} else {
      LBPt=TLin; LBPm=MLin; LBPb=BLin;
	}
    // copy line into buffer
    if(y<(Height-1)){
      memcpy(LBPb+1,bitptr1,ScLen);
      // extend borders
      LBPb[0]=LBPb[1]; LBPb[Width+1]=LBPb[Width];
    }
    for(x=0; x<Width; x++){
	  cr = BlendRGB4(LBPt[x+1],LBPm[x  ],LBPm[x+2],LBPb[x+1]);
	  dg = BlendRGB4(LBPt[x  ],LBPt[x+2],LBPb[x  ],LBPb[x+2]);
	  ((unsigned long int *)bitptr2)[x] = BlendRGB2(LBPm[x+1], BlendRGB3(cr, cr, dg));
    }
    bitptr2 = bitptr1; bitptr1 += ScLen;
  }
}

vector   VPorg1, XD, YD;
unsigned long int *fbf;
int      XRs1,  cd;

unsigned long int cc,bc,cr,br;

unsigned long int DoSubAA(float x, float y, unsigned long int br, unsigned long int tl, unsigned long int bl, unsigned long int tr)
{
  cd = DiffRGB4(br, tr, bl, tl);
  if(cd>20){
    unsigned long int cc,bc,cr;
    cc =   Clamp2ABGR(RayTrace0(unit(VPorg1 + (y-0.25f)*YD + (x-0.25f)*XD) ));
    cd =   DiffRGB2(br, bl);
    if(cd>20){
	  bc = Clamp2ABGR(RayTrace0(unit(VPorg1 + (y)      *YD + (x-0.25f)*XD) ));
	} else {
      bc = BlendRGB2(br, bl);
	}
    cd = DiffRGB2(br, tr);
    if(cd>20){
	  cr = Clamp2ABGR(RayTrace0(unit(VPorg1 + (y-0.25f)*YD + (x)      *XD) ));
	} else {
      cr = BlendRGB2(br, tr);
	}
    return BlendRGB4(br,cc,bc,cr);
  } else {
	return br;
  }
}

void DoAntiAlias(int x, int y)
{
  cd = DiffRGB4(*(fbf+(y)*XRs1+(x)), *(fbf+(y-1)*XRs1+(x)), *(fbf+(y)*XRs1+(x-1)), *(fbf+(y-1)*XRs1+(x-1)) );
  if(cd>20){
	cc =   Clamp2ABGR(RayTrace0(unit(VPorg1 + (((float)y)-0.5f)*YD + (((float)x)-0.5f)*XD) ));
    cd =   DiffRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y  )*XRs1+(x-1)) );
    if(cd>20){
	  bc = Clamp2ABGR(RayTrace0(unit(VPorg1 +  ((float)y)      *YD + (((float)x)-0.5f)*XD) ));
	} else {
      bc = BlendRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y  )*XRs1+(x-1)) );
	}
    cd = DiffRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y-1)*XRs1+(x  )) );
    if(cd>20){
	  cr = Clamp2ABGR(RayTrace0(unit(VPorg1 + (((float)y)-0.5f)*YD +  ((float)x)      *XD) ));
	} else {
      cr = BlendRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y-1)*XRs1+(x  )) );
	}
    //br = *(fbf+y*XRs1+x);
    br = DoSubAA(((float)x)     , ((float)y)     , *(fbf+y*XRs1+x), cc, bc, cr);
	bc = DoSubAA(((float)x)-0.5f, ((float)y)     , bc, BlendRGB2(*(fbf+(y  )*XRs1+(x-1)), *(fbf+(y-1)*XRs1+(x-1))), *(fbf+(y  )*XRs1+(x-1)), cc );
	cr = DoSubAA(((float)x)     , ((float)y)-0.5f, cr, BlendRGB2(*(fbf+(y-1)*XRs1+(x  )), *(fbf+(y-1)*XRs1+(x-1))), cc, *(fbf+(y-1)*XRs1+(x  )) );
	cc = DoSubAA(((float)x)-0.5f, ((float)y)-0.5f, cc, *(fbf+(y-1)*XRs1+(x-1)), BlendRGB2(*(fbf+(y  )*XRs1+(x-1)), *(fbf+(y-1)*XRs1+(x-1))), BlendRGB2(*(fbf+(y-1)*XRs1+(x  )), *(fbf+(y-1)*XRs1+(x-1))) );
    *(fbf+y*XRs1+x) = BlendRGB4(br,cc,bc,cr);
  }
}

void Quadrant2(int x, int y, char q)
{
  char q1,q2;
  if(!(q & 1)){
    cd = DiffRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y  )*XRs1+(x-2)) );
    if(cd>80){
      *(fbf+(y  )*XRs1+(x-1)) = Clamp2ABGR(RayTrace0(unit(VPorg1 + ((float)(y  ))*YD + ((float)(x-1))*XD) )); q1=1;
	} else {
      *(fbf+(y  )*XRs1+(x-1)) = BlendRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y  )*XRs1+(x-2)) ); q1=0;
	}
  }
  if(!(q & 2)){
    cd = DiffRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y-2)*XRs1+(x  )) );
    if(cd>80){
      *(fbf+(y-1)*XRs1+(x  )) = Clamp2ABGR(RayTrace0(unit(VPorg1 + ((float)(y-1))*YD + ((float)(x  ))*XD) )); q2=1;
	} else {
      *(fbf+(y-1)*XRs1+(x  )) = BlendRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y-2)*XRs1+(x  )) ); q2=0;
	}
  }
  cd = max(DiffRGB2(*(fbf+(y-1)*XRs1+(x  )), *(fbf+(y-1)*XRs1+(x-2))),
	       DiffRGB2(*(fbf+(y  )*XRs1+(x-1)), *(fbf+(y-2)*XRs1+(x-1))) );
  if(cd>80){
    *(fbf+(y-1)*XRs1+(x-1)) = Clamp2ABGR(RayTrace0(unit(VPorg1 + ((float)(y-1))*YD + ((float)(x-1))*XD) ));
  } else {
    *(fbf+(y-1)*XRs1+(x-1)) = BlendRGB4(*(fbf+(y-1)*XRs1+(x  )), *(fbf+(y-1)*XRs1+(x-2)), *(fbf+(y  )*XRs1+(x-1)), *(fbf+(y-2)*XRs1+(x-1)) );
  }
}

void RV_RenderImage(byte *BMPbits, int XRes, int YRes, int border)
{
  int        x,y;
  RGB_Color  Col;
  vector     Ray;

  fbf = (unsigned long int *)BMPbits; XRs1=XRes+border; 

  vector N  = unit(RV_Target - *RV_CamPtr);
  vector V  = unit(RV_Up-(RV_Up%N)*N);
  vector U  = N*V;

  float  b  = (float)tan(RV_FoV/2.0f);
  float  bx = XRes<YRes? b : b*((float)XRes/(float)YRes)    ;
  float  by = XRes<YRes?     b*((float)YRes/(float)XRes) : b;

  VPorg1 = N - bx*U - by*V;
  YD     = 2.0f/(float)YRes*by*V;
  XD     = 2.0f/(float)XRes*bx*U;
  
  if(RV_Quality==1){
    for(y=0; y<YRes; y++){
	  Ray = VPorg1 + ((float)y)*YD;
      for(x=border; x<XRs1; x++){
		/*if((x==XRs1/7) && (y==YRes/2)){
		  Col=RGB_Color(0.0f, 0.0f, 0.0f);
		}*/
		Col = RayTrace0(unit(Ray + ((float)x)*XD));
	    *(fbf+y*XRs1+x)=Clamp2ABGR(Col);
	  }
	}
  } else if(RV_Quality==4){
	RV_MDP[0]=((float *)RV_CamPtr)[0]; RV_MDP[1]=((float *)RV_CamPtr)[1]; RV_MDP[2]=((float *)RV_CamPtr)[2]; RV_MDP[3]=1.0f;
    for(y=0; y<YRes; y+=2){
      for(x=border; x<XRs1; x+=2){
		Ray = unit(VPorg1 + ((float)y  )*YD + ((float)x  )*XD);
	    RV_MDP[16]=Ray.x; RV_MDP[20]=Ray.y; RV_MDP[24]=Ray.z;
		Ray = unit(VPorg1 + ((float)y  )*YD + ((float)x+1)*XD);
	    RV_MDP[17]=Ray.x; RV_MDP[21]=Ray.y; RV_MDP[25]=Ray.z;
		Ray = unit(VPorg1 + ((float)y+1)*YD + ((float)x  )*XD);
	    RV_MDP[18]=Ray.x; RV_MDP[22]=Ray.y; RV_MDP[26]=Ray.z;
		Ray = unit(VPorg1 + ((float)y+1)*YD + ((float)x+1)*XD);
	    RV_MDP[19]=Ray.x; RV_MDP[23]=Ray.y; RV_MDP[27]=Ray.z;
		RayTrace4();
	    *(fbf+ y   *XRs1+x  )=Clamp2ABGR(RGB_Color(RV_MDP[84],RV_MDP[88],RV_MDP[92]));
	    *(fbf+ y   *XRs1+x+1)=Clamp2ABGR(RGB_Color(RV_MDP[85],RV_MDP[89],RV_MDP[93]));
	    *(fbf+(y+1)*XRs1+x  )=Clamp2ABGR(RGB_Color(RV_MDP[86],RV_MDP[90],RV_MDP[94]));
	    *(fbf+(y+1)*XRs1+x+1)=Clamp2ABGR(RGB_Color(RV_MDP[87],RV_MDP[91],RV_MDP[95]));
	  }
	}
  } else if(RV_Quality==2){
    Col = RayTrace0(unit(VPorg1)); *(fbf)=Clamp2ABGR(Col); char q00,q01,q10,q11;
	for(x=4; x<XRes+1; x+=4){
      Col = RayTrace0(unit(VPorg1 + ((float)x)*XD));
	  *(fbf+ x   )=Clamp2ABGR(Col);
      *(fbf+(x-2))=BlendRGB2(*(fbf+(x  )), *(fbf+(x-4)) );
      *(fbf+(x-1))=BlendRGB2(*(fbf+(x  )), *(fbf+(x-2)) );
      *(fbf+(x-3))=BlendRGB2(*(fbf+(x-2)), *(fbf+(x-4)) );
    }
	for(y=4; y<YRes+1; y+=4){
      Col = RayTrace0(unit(VPorg1 + ((float)y)*YD));
	  *(fbf+ y   *XRs1)=Clamp2ABGR(Col);
      *(fbf+(y-2)*XRs1)=BlendRGB2(*(fbf+(y  )*XRs1), *(fbf+(y-4)*XRs1) );
      *(fbf+(y-1)*XRs1)=BlendRGB2(*(fbf+(y  )*XRs1), *(fbf+(y-2)*XRs1) );
      *(fbf+(y-3)*XRs1)=BlendRGB2(*(fbf+(y-2)*XRs1), *(fbf+(y-4)*XRs1) );
	}
	for(y=4; y<YRes+1; y+=4){
	  for(x=4; x<XRes+1; x+=4){
	    *(fbf+y*XRs1+x) = Clamp2ABGR(RayTrace0(unit(VPorg1 + ((float)y)*YD + ((float)x)*XD) ));
        cd = DiffRGB4(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y-4)*XRs1+(x-4)), *(fbf+(y  )*XRs1+(x-4)), *(fbf+(y-4)*XRs1+(x  )) );
		if(abs(cd)>5){
		  q00=0; q01=0; q10=0; q11=0;
          cd = DiffRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y  )*XRs1+(x-4)) );
          if(cd>60){
		    *(fbf+(y  )*XRs1+(x-2)) = Clamp2ABGR(RayTrace0(unit(VPorg1 + ((float)(y  ))*YD + ((float)(x-2))*XD) ));
		  } else {
            *(fbf+(y  )*XRs1+(x-2)) = BlendRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y  )*XRs1+(x-4)) );
            *(fbf+(y  )*XRs1+(x-1)) = BlendRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y  )*XRs1+(x-2)) );
            *(fbf+(y  )*XRs1+(x-3)) = BlendRGB2(*(fbf+(y  )*XRs1+(x-2)), *(fbf+(y  )*XRs1+(x-4)) );
			q00|=1; q10|=1;
		  }
          cd = DiffRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y-4)*XRs1+(x  )) );
          if(cd>60){
		    *(fbf+(y-2)*XRs1+(x  )) = Clamp2ABGR(RayTrace0(unit(VPorg1 + ((float)(y-2))*YD + ((float)(x  ))*XD) ));
		  } else {
            *(fbf+(y-2)*XRs1+(x  )) = BlendRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y-4)*XRs1+(x  )) );
            *(fbf+(y-1)*XRs1+(x  )) = BlendRGB2(*(fbf+(y  )*XRs1+(x  )), *(fbf+(y-2)*XRs1+(x  )) );
            *(fbf+(y-3)*XRs1+(x  )) = BlendRGB2(*(fbf+(y-2)*XRs1+(x  )), *(fbf+(y-4)*XRs1+(x  )) );
			q00|=2; q01|=2;
		  }
          cd = max(DiffRGB2(*(fbf+(y-2)*XRs1+(x  )), *(fbf+(y-2)*XRs1+(x-4))),
			       DiffRGB2(*(fbf+(y  )*XRs1+(x-2)), *(fbf+(y-4)*XRs1+(x-2))) );
          if(cd>60){
		    *(fbf+(y-2)*XRs1+(x-2)) = Clamp2ABGR(RayTrace0(unit(VPorg1 + ((float)(y-2))*YD + ((float)(x-2))*XD) ));
		  } else {
            *(fbf+(y-2)*XRs1+(x-2)) = BlendRGB4(*(fbf+(y-2)*XRs1+(x  )), *(fbf+(y-2)*XRs1+(x-4)), *(fbf+(y  )*XRs1+(x-2)), *(fbf+(y-4)*XRs1+(x-2)) );
            *(fbf+(y-2)*XRs1+(x-1)) = BlendRGB2(*(fbf+(y-2)*XRs1+(x  )), *(fbf+(y-2)*XRs1+(x-2)) );
            *(fbf+(y-2)*XRs1+(x-3)) = BlendRGB2(*(fbf+(y-2)*XRs1+(x-2)), *(fbf+(y-2)*XRs1+(x-4)) );
            *(fbf+(y-3)*XRs1+(x-2)) = BlendRGB2(*(fbf+(y-2)*XRs1+(x-2)), *(fbf+(y-4)*XRs1+(x-2)) );
            *(fbf+(y-1)*XRs1+(x-2)) = BlendRGB2(*(fbf+(y  )*XRs1+(x-2)), *(fbf+(y-2)*XRs1+(x-2)) );
			q10|=2; q01|=1; q11|=3;
		  }
          Quadrant2(x-2,y-2,q11); Quadrant2(x  ,y-2,q01); Quadrant2(x-2,y  ,q10); Quadrant2(x  ,y  ,q00);
		} else {
		  unsigned long int *pixels;
		  unsigned long int  color = *(fbf+y*XRs1+x);
		  for(int yi=0; yi<4; yi++){
		    pixels=fbf+(y-yi)*XRs1;
			if(yi) pixels[x  ]=color;
			       pixels[x-1]=color;
			       pixels[x-2]=color;
			       pixels[x-3]=color;
		  }
		}
      }
	}
  }
  if(RV_AntiAlias){
	for(y=1; y<YRes; y++){
	  for(x=1; x<XRes; x++){
		DoAntiAlias(x,y);
	  }
	}
  }
  if(RV_Filter) DoFilter(BMPbits, XRes, YRes);
}
